__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2019-12-11'


import logging
import os

import numpy as np
import pandas as pd
import xarray as xr
import statsmodels.api as sm
import keras

from src.run_modules.run_environment import RunEnvironment
from src.data_handling.data_distributor import Distributor
from src.data_handling.data_generator import DataGenerator
from src.model_modules.linear_model import OrdinaryLeastSquaredModel
from src import statistics
from src import helpers
from src.helpers import TimeTracking
from src.plotting.postprocessing_plotting import plot_monthly_summary, plot_climsum_boxplot, plot_station_map, plot_conditional_quantiles
from src.datastore import NameNotFoundInDataStore


class PostProcessing(RunEnvironment):

    def __init__(self):
        super().__init__()
        self.model: keras.Model = self._load_model()
        self.ols_model = None
        self.batch_size: int = self.data_store.get("batch_size", "general.model")
        self.test_data: DataGenerator = self.data_store.get("generator", "general.test")
        self.test_data_distributed = Distributor(self.test_data, self.model, self.batch_size)
        self.train_data: DataGenerator = self.data_store.get("generator", "general.train")
        self.plot_path: str = self.data_store.get("plot_path", "general")
        self._run()

    def _run(self):
        self.train_ols_model()
        preds_for_all_stations = self.make_prediction()
        self.plot()

    def _load_model(self):
        try:
            model = self.data_store.get("best_model", "general")
        except NameNotFoundInDataStore:
            logging.info("no model saved in data store. trying to load model from experiment")
            path = self.data_store.get("experiment_path", "general")
            name = f"{self.data_store.get('experiment_name', 'general')}_my_model.h5"
            model_name = os.path.join(path, name)
            model = keras.models.load_model(model_name)
        return model

    def plot(self):
        logging.debug("Run plotting routines...")
        path = self.data_store.get("forecast_path", "general")
        window_lead_time = self.data_store.get("window_lead_time", "general")
        target_var = self.data_store.get("target_var", "general")

        plot_conditional_quantiles(self.test_data.stations, plot_folder=self.plot_path, pred_name="CNN", ref_name="orig",
                                   forecast_path=self.data_store.get("forecast_path", "general"), plot_name_affix="cali-ref")
        plot_conditional_quantiles(self.test_data.stations, plot_folder=self.plot_path, pred_name="orig", ref_name="CNN",
                                   forecast_path=self.data_store.get("forecast_path", "general"), plot_name_affix="like-bas")
        plot_station_map(generators={'b': self.test_data}, plot_folder=self.plot_path)
        plot_monthly_summary(self.test_data.stations, path, r"forecasts_%s_test.nc", window_lead_time, target_var,
                             plot_folder=self.plot_path)
        # plot_climsum_boxplot()

    def calculate_test_score(self):
        test_score = self.model.evaluate_generator(generator=self.test_data_distributed.distribute_on_batches(),
                                                   use_multiprocessing=False, verbose=0, steps=1)
        logging.info(f"test score = {test_score}")
        self._save_test_score(test_score)

    def _save_test_score(self, score):
        path = self.data_store.get("experiment_path", "general")
        with open(os.path.join(path, "test_scores.txt")) as f:
            for index, item in enumerate(score):
                f.write(f"{self.model.metrics[index]}, {item}\n")

    def train_ols_model(self):
        self.ols_model = OrdinaryLeastSquaredModel(self.train_data)

    def make_prediction(self, freq="1D"):
        logging.debug("start make_prediction")
        nn_prediction_all_stations = []
        for i, v in enumerate(self.test_data):
            data = self.test_data.get_data_generator(i)

            nn_prediction, persistence_prediction, ols_prediction = self._create_empty_prediction_arrays(data, count=3)
            input_data = self.test_data[i][0]

            # get scaling parameters
            mean, std, transformation_method = data.get_transformation_information(variable='o3')

            # nn forecast
            nn_prediction = self._create_nn_forecast(input_data, nn_prediction, mean, std, transformation_method)

            # persistence
            persistence_prediction = self._create_persistence_forecast(input_data, persistence_prediction, mean, std, 
                                                                       transformation_method)

            # ols
            ols_prediction = self._create_ols_forecast(input_data, ols_prediction, mean, std, transformation_method)

            # orig pred
            orig_pred = self._create_orig_forecast(data, None, mean, std, transformation_method)

            # merge all predictions
            full_index = self.create_fullindex(data.data.indexes['datetime'], freq)
            all_predictions = self.create_forecast_arrays(full_index, list(data.label.indexes['window']),
                                                          CNN=nn_prediction,
                                                          persi=persistence_prediction,
                                                          orig=orig_pred,
                                                          OLS=ols_prediction)

            # save all forecasts locally
            path = self.data_store.get("forecast_path", "general")
            file = os.path.join(path, f"forecasts_{data.station[0]}_test.nc")
            all_predictions.to_netcdf(file)

            # save nn forecast to return variable
            nn_prediction_all_stations.append(nn_prediction)
        return nn_prediction_all_stations

    @staticmethod
    def _create_orig_forecast(data, placeholder, mean, std, transformation_method):
        return statistics.apply_inverse_transformation(data.label, mean, std, transformation_method)

    def _create_ols_forecast(self, input_data, ols_prediction, mean, std, transformation_method):
        tmp_ols = self.ols_model.predict(input_data)
        tmp_ols = statistics.apply_inverse_transformation(tmp_ols, mean, std, transformation_method)
        ols_prediction.values = np.swapaxes(np.expand_dims(tmp_ols, axis=1), 2, 0)
        return ols_prediction

    def _create_persistence_forecast(self, input_data, persistence_prediction, mean, std, transformation_method):
        tmp_persi = input_data.sel({'window': 0, 'variables': 'o3'})
        tmp_persi = statistics.apply_inverse_transformation(tmp_persi, mean, std, transformation_method)
        window_lead_time = self.data_store.get("window_lead_time", "general")
        persistence_prediction.values = np.expand_dims(np.tile(tmp_persi.squeeze('Stations'), (window_lead_time, 1)),
                                                       axis=1)
        return persistence_prediction

    def _create_nn_forecast(self, input_data, nn_prediction, mean, std, transformation_method):
        tmp_nn = self.model.predict(input_data)
        tmp_nn = statistics.apply_inverse_transformation(tmp_nn, mean, std, transformation_method)
        nn_prediction.values = np.swapaxes(np.expand_dims(tmp_nn, axis=1), 2, 0)
        return nn_prediction

    @staticmethod
    def _create_empty_prediction_arrays(generator, count=1):
        return [generator.label.copy()] * count

    @staticmethod
    def create_fullindex(df, freq):
        # Diese Funkton erstellt ein leeres df, mit Index der Frequenz frequ zwischen dem ersten und dem letzten Datum in df
        # param: df as pandas dataframe
        # param: freq as string
        # return: index as pandas dataframe
        if isinstance(df, pd.DataFrame):
            earliest = df.index[0]
            latest = df.index[-1]
        elif isinstance(df, xr.DataArray):
            earliest = df.index[0].values
            latest = df.index[-1].values
        elif isinstance(df, pd.core.indexes.datetimes.DatetimeIndex):
            earliest = df[0]
            latest = df[-1]
        else:
            raise AttributeError(f"unknown array type. Only pandas dataframes, xarray dataarrays and pandas datetimes "
                                 f"are supported. Given type is {type(df)}.")
        index = pd.DataFrame(index=pd.date_range(earliest, latest, freq=freq))
        return index

    @staticmethod
    def create_forecast_arrays(index, ahead_names, **kwargs):
        """
        This function combines different forecast types into one xarray.

        :param index: as index; index for forecasts (e.g. time)
        :param ahead_names: as list of str/int: names of ahead values (e.g. hours or days)
        :param kwargs: as xarrays; data of forecasts
        :return: xarray of dimension 3: index, ahead_names, # predictions

        """
        keys = list(kwargs.keys())
        res = xr.DataArray(np.full((len(index.index), len(ahead_names), len(keys)), np.nan),
                           coords=[index.index, ahead_names, keys], dims=['index', 'ahead', 'type'])
        for k, v in kwargs.items():
            try:
                match_index = np.stack(set(res.index.values) & set(v.index.values))
                res.loc[match_index, :, k] = v.loc[match_index]
            except AttributeError:  # v is xarray type and has no attribute .index
                match_index = np.stack(set(res.index.values) & set(v.indexes['datetime'].values))
                res.loc[match_index, :, k] = v.sel({'datetime': match_index}).squeeze('Stations').transpose()
        return res


